// PixMap.java

/*
 * STATUS
 *
 * Remove runtime exceptions
 */

package uk.org.redfelineninja.awt;

/*
 * I have been rather loose with Americanisms (esp. Color vs Colour). My
 * comments usually talk about Colour, my code (hopefully) talks about Color
 * as this saves confusion with standard librarys
 */

import uk.org.redfelineninja.util.IntStack;

import java.awt.*;
import java.awt.image.*;

/**
 * A component similar to Canvas but with per pixel access built in.
 *
 * This class originated through dissatisfaction with java.awt.Graphics. I need
 * to test the colour of a pixel. java.awt.Graphics does not seem to provide
 * this.
 */
public class PixMap extends Component {
    // DATA MEMBERS

    // data (let sub-classers do horrific things to my heart!)

    /**
     * The pixels of the image in int format. Sub classes are free to play
     * with these.
     */
    protected int pixels[];

    /**
     * The size of the image. A change in image size really ought to mean
     * a new PixMap so this has been made private.
     * Sorry.
     */
    private Dimension size;
    ColorModel myColorModel;

    // state
    int color = 0xff000000;
    boolean autoRepaint = true;

    // derived
    MemoryImageSource imageSource;
    Image myImage;

    // utility
    MediaTracker tracker;

    /**
     * Create a new PixMap from an Image using a direct RGB colour mapping.
     *
     * @param	img	Image from which to generate the PixMap
     */
    public PixMap(Image img) {
        this(img, ColorModel.getRGBdefault());
    } // PixMap( Image )

    /**
     * Create a PixMap from an Image using the supplied ColorModel.
     *
     * @param	img	Image from which to generate the PixMap
     * @param	cm	ColorModel to apply to the image
     * @see java.awt.image.ColorModel
     */
    public PixMap(Image img, ColorModel cm) {
        // simple assigment
        myColorModel = cm;

        // ensure the image is loaded
        // we are in a contructor so can't pass this to tracker
        Component bombProof = new Component() {};
        tracker = new MediaTracker(bombProof);
        tracker.addImage(img, 0);
        try {
            tracker.waitForID(0);
        } catch (InterruptedException e) {
            // this is unlikely so we stack dump
            e.printStackTrace();
        }

        if (tracker.isErrorAny()) {
            // can't load image
            pixels = null;
            size = new Dimension(0, 0);
            throw new RuntimeException("Can't load image");
        }

        // these should be available now
        size = new Dimension(img.getWidth(this), img.getHeight(this));

        // get the pixels
        pixels = new int[size.width * size.height];
        PixelGrabber pg = new PixelGrabber(img, 0, 0, size.width, size.height, pixels, 0, size.width);

        try {
            pg.grabPixels();
        } catch (InterruptedException e) {
            // this is unlikely so we stack dump
            e.printStackTrace();
        }

        // derived data members
        imageSource = new MemoryImageSource(size.width, size.height, myColorModel, pixels, 0, size.width);
        imageSource.setAnimated(true);
        myImage = createImage(imageSource);

        // set up media tracker
        tracker.removeImage(img, 0);
        tracker.addImage(myImage, 0);
    } // PixMap( Image, ColorModel )

    /**
     * Supplies the Image version of the PixMap.
     *
     * Provides a means to use the graphics operations of PixMap without
     * having to build the Component into a GUI.
     *
     * @return The image version of the PixMap
     */
    public Image getImage() {
        return myImage;
    }

    /**
     * Supplies the pixels used in this PixMap.
     *
     * Provides a means for direct manipulation of pixels. Nine times out
     * of ten it would be better to sub-class a PixMap and make use of the
     * direct manipulation provided by that route.
     *
     * @return The array of integers that the PixMap is generated from
     */
    public final int[] getPixels() {
        return pixels;
    }

    /**
     * Get the colour of the pixel at (x, y).
     *
     * This method is not clever and so you may get an ArrayIndexOutofBoundsException thrown to you.
     *
     * @return Integer representing a colour. Bits 24-31 represent alpha, 16-23, red, 8-15 green and 0-7 blue.
     */
    public final int getPixel(int x, int y) {
        return pixels[x + y * size.width];
    }

    /**
     * Get the colour of the pixel at the supplied point.
     *
     * This method is not clever and so you may get an ArrayIndexOutofBoundsException thrown to you.
     *
     * @return Integer representing a colour. Bits 24-31 represent alpha, 16-23, red, 8-15 green and 0-7 blue.
     */
    public final int getPixel(Point d) {
        return pixels[d.x + d.y * size.width];
    }

    /**
     * Set the colour of the pixel at (x, y).
     *
     * This method is not clever and so you may get an ArrayIndexOutofBoundsException thrown to you.
     *
     * @param color Integer representing a colour. Bits 24-31 represent alpha, 16-23, red, 8-15 green and 0-7 blue.
     */
    public final void setPixel(int x, int y, int color) {
        pixels[x + y * size.width] = color;
    }

    /**
     * Set the colour of the pixel at the supplied point.
     *
     * This method is not clever and so you may get an ArrayIndexOutofBoundsException thrown to you.
     *
     * @param color Integer representing a colour. Bits 24-31 represent alpha, 16-23, red, 8-15 green and 0-7 blue.
     */
    public final void setPixel(Point d, int color) {
        pixels[d.x + d.y * size.width] = color;
    }

    /**
     * Returns the ColorModel used when creating the image from the PixMap
     *
     * @return the ColorModel used in this PixMap
     */
    public ColorModel getColorModel() {
        return myColorModel;
    }

    /**
     * Sets the ColorModel used when creating the image from the PixMap
     */
    public void setColorModel(ColorModel cm) {
        myColorModel = cm;
        imageSource.newPixels(pixels, myColorModel, 0, size.width);
    }

    /**
     * Get the current draw colour.
     *
     * @return Integer representing a colour. Bits 24-31 represent alpha, 16-23, red, 8-15 green and 0-7 blue.
     */
    public int getColor() {
        return color;
    }

    /**
     * Sets the draw colour.
     *
     * @param c Integer representing a colour. Bits 24-31 represent alpha, 16-23, red, 8-15 green and 0-7 blue.
     */
    public void setColor(int c) {
        color = c;
    }

    /**
     * Sets the draw colour.
     */
    public void setColor(Color c) {
        color = c.getRGB();
    }

    /**
     * Draws a line interior to the supplied points.
     *
     * This method using bresenhams algorithm draw a line between two points.
     */
    public void drawLine(int x1, int y1, int x2, int y2) {
        /*
         * This method is where possible hand optimised in case of poor compilers
         * this may make it harder to read.
         */

        int dx = x2 - x1;
        int dy = y2 - y1;

        // check for trivial solutions
        if (dx == 0) {
            // vertical line
            for (int y = y1; y < y2; y++) {
                pixels[x1 + y * size.width] = color;
            }

            return;
        } else if (dy == 0) {
            // horizontal line
            for (int x = x1; x < x2; x++) {
                pixels[x + y1 * size.width] = color;
            }

            return;
        } // if( trivial solution )

        // initialise
        int x = x1;
        int y = y1;
        pixels[x + y * size.width] = color;

        // bresenham constant negative
        int bkn = dy + dy;

        // bresenham constant positve
        int bkp = bkn - dx - dx;

        // decision variable
        int p = dy + bkp;

        // main loop (theroreticall from 0 to dx - 2
        for (int k = 1; k < dx; k++) {
            if (p < 0) {
                p += bkn;
            } else {
                p += bkp;
                y++;
            }

            x++;
            pixels[x + y * size.width] = color;
        } // main loop

        if (autoRepaint) {
            repaint(x1, y1, x2 - 1, y2 - 1);
        }
    } // drawLine( int, int, int, int )

    /**
     * Get the preferred size of this component.
     *
     * This is the actual size of the PixMap and can be used to calculate
     * indexs if the pixels are accessed directly.
     *
     * @return the size of the PixMap
     * @see #getPixels()
     */
    public Dimension getPreferredSize() {
        return size;
    }

    /** 
     * Get the minimum size of this component.
     *
     * This is the actual size of the PixMap and can be used to calculate
     * indexs if the pixels are accessed directly.
     *
     * @return the size of the PixMap
     * @see #getPixels()
     */
    public Dimension getMinimumSize() {
        return size;
    }

    /** 
     * Get the maximum size of this component.
     *
     * This is the actual size of the PixMap and can be used to calculate
     * indexs if the pixels are accessed directly.
     *
     * @return the size of the PixMap
     * @see #getPixels()
     */
    public Dimension getMaximumSize() {
        return size;
    }

    /**
     * Find whether or not the PixMap will automatically repaint itself.
     *
     * @return true if the PixMap will automatically repaint.
     */
    public boolean getAutoRepaint() {
        return autoRepaint;
    }

    /**
     * Set whether or not the PixMap will automatically repaint itself.
     *
     * @param ar true if the PixMap should automatically repaint.
     */
    public void setAutoRepaint(boolean ar) {
        autoRepaint = ar;
    }

    /**
     * Uses a simple recursive algorithm to perform a flood fill.
     *
     * <p><strong>WARNING</strong>
     * <br>The algorithm to perform flood filling is deeply recursive. It works
     * fine on most virtual machines. It only works for very small areas
     * on PersonalJava VMs, it throws a <tt>StackOverflowException</tt> if
     * it does not work. It will be updated when a better way comes to light.
     *
     * @return the bounding box of the flood fill.
     */
    public Rectangle floodFill(int x, int y) {
        int oldColor = pixels[x + y * size.width];
        Rectangle bounds = new Rectangle(x, y, 1, 1);

        IntStack stackX = new IntStack();
        IntStack stackY = new IntStack();

        // be very suspect of an op on stackX without a matching one on stackY
        stackX.push(x);
        stackY.push(y);

        while (!stackX.empty()) {
            x = stackX.pop();
            y = stackY.pop();

            pixels[x + y * size.width] = color;
            bounds.add(x, y);

            // up ( forces lazy evaluation for correctness)
            if (y != 0)
                if (pixels[x + (y - 1) * size.width] == oldColor) {
                    stackX.push(x);
                    stackY.push(y - 1);
                } // if( up )

            // down (forces lazy evaluation for correctness)
            if (y != size.height)
                if (pixels[x + (y + 1) * size.width] == oldColor) {
                    stackX.push(x);
                    stackY.push(y + 1);
                } // if( down )

            // left (forces lazy evaluation for correctness)
            if (x != 0)
                if (pixels[(x - 1) + y * size.width] == oldColor) {
                    stackX.push(x - 1);
                    stackY.push(y);
                } // if( left )

            // right (forces lazy evaluation for correctness)
            if (x != size.width)
                if (pixels[(x + 1) + y * size.width] == oldColor) {
                    stackX.push(x + 1);
                    stackY.push(y);
                } // if( right )

        } // while( not empty )

        // in java the width and height are EXCLUSIVE
        imageSource.newPixels(bounds.x, bounds.y, bounds.width + 1, bounds.height + 1);

        if (autoRepaint)
            repaint(bounds.x, bounds.y, bounds.width + 1, bounds.height + 1);

        return bounds;

    } // floodFill( int, int )

    public Rectangle floodFillOld(int x, int y) {
        int oldColor = pixels[x + y * size.width];
        Rectangle r = new Rectangle(x, y, 1, 1);

        r = floodFillPeer(x, y, oldColor, r);

        // in java the width and height are EXCLUSIVE
        imageSource.newPixels(r.x, r.y, r.width + 1, r.height + 1);

        if (autoRepaint)
            repaint(r.x, r.y, r.width + 1, r.height + 1);

        return r;
    } // floodFill( int, int, int )

    /*
     * Peer to floodFill. Could be made more effiecent by spilting into 4
     * methods up, down left and right and not checking behind us or by using
     * 'clipboard' variables (clipboards would break the thread safe design of
     * this method)
     */
    private final Rectangle floodFillPeer(int x, int y, int oldColor, Rectangle bounds) {
        pixels[x + y * size.width] = color;
        bounds.add(x, y);

        // up (forces lazy evaluation for correctness)
        if (y != 0)
            if (pixels[x + (y - 1) * size.width] == oldColor)
                bounds = floodFillPeer(x, y - 1, oldColor, bounds);

        // down (forces lazy evaluation for correctness)
        if (y != size.height)
            if (pixels[x + (y + 1) * size.width] == oldColor)
                bounds = floodFillPeer(x, y + 1, oldColor, bounds);

        // left (forces lazy evaluation for correctness)
        if (x != 0)
            if (pixels[(x - 1) + y * size.width] == oldColor)
                bounds = floodFillPeer(x - 1, y, oldColor, bounds);

        // right (forces lazy evaluation for correctness)
        if (x != size.width)
            if (pixels[(x + 1) + y * size.width] == oldColor)
                bounds = floodFillPeer(x + 1, y, oldColor, bounds);

        return bounds;
    } // floodFillPeer()

    /**
     * Tells the PixMap that the whole of is pixels need to be updated.
     */
    public void newPixels() {
        imageSource.newPixels();
        if (autoRepaint)
            repaint();
    }

    /**
     * Tells the PixMap that the pixels within the supplied box need to be
     * updated.
     *
     * @param x the x coordinate of the upper left corner of the new Pixels
     * @param y the y coordinate of the upper left corner of the new Pixels
     * @param w the width of the rectangle
     * @param h the height of the rectangle
     */
    public void newPixels(int x, int y, int w, int h) {
        imageSource.newPixels(x, y, w, h);
        if (autoRepaint)
            repaint(x, y, w, h);
    }

    /**
     * Ensures the image is ready to be output.
     *
     * This is only really for the paranoid, since the image is sourced from
     * memory there is no reason for in not to be loaded.
     *
     * @return	false if the image could not be loaded
     */
    public boolean loadImage() {
        try {
            tracker.waitForAll();
        } catch (InterruptedException e) {
            // this is unlikely so we stack dump
            e.printStackTrace();
        }

        return !tracker.isErrorAny();
    }

    // THESE OVERRIDE METHODS OF COMPONENT

    /**
     * Draw the Image as supplied by getImage to screen.
     *
     * @see #getImage()
     */
    public void paint(Graphics g) {
        g.drawImage(myImage, 0, 0, this);
    } // paint( Graphics )

} // PixMap
