// AbstractMulticaster.java

package uk.org.redfelineninja.net;

import java.util.Vector;
import java.lang.reflect.Method;
import java.lang.reflect.InvocationTargetException;

/**
 * This class provides a rapid means to develop any multicaster. The meaning of
 * multicaster in this sense means a class that acts like a single class and
 * passes all method calls on itself to every known listener.
 * <p>
 * A new multicaster should have methods that look something like this example
 * <pre>
 * public void message( String foo, int bar ) {
 *    Object [] args = new Object[ 2 ];
 * 
 *    // copy args in
 *    args[0] = foo;
 *    args[1] = new Integer( bar );
 * 
 *    // invoke the methods
 *    invokeMethod( "message", args );
 * }
 * </pre>
 *
 * @see GenericMulticaster
 */
public abstract class AbstractMulticaster {
    // this is the list of clients for this Multicaster
    private Vector clients;

    /**
     * Create a new multicaster with no listeners attached.
     */
    public AbstractMulticaster() {
        clients = new Vector();
    }

    /**
     * Add a listener to this multicaster.
     */
    public void addListener(Object listener) {
        clients.addElement(listener);
    }

    /**
     * Remove a listener from this multicaster.
     */
    public void removeListener(Object listener) {
        clients.removeElement(listener);
    }

    /**
     * For each known client attempt to call the supplied method using the
     * supplied list of arguments.
     * <p>
     * WARNING: If the argument is a primative wrapper (eg java.lang.Integer) we
     * do not know if it represents an int or java.lang.Integer, we assume in
     * this method that it is an int.
     *
     * @see		GenericMulticaster
     *
     * @param	name	the name of the method that should be called
     * @param	args	the list of arguments to be supplied to the method
     *
     * @return		an array of Objects
     */
    protected Object[] invokeMethod(String name, Object[] args) {
        // copy the clients out of the Vector
        Object[] clnts = new Object[clients.size()];
        clients.copyInto(clnts);

        // get the argList ready
        Class[] argList;

        if (args == null) {
            // no arguments implies to argument list
            argList = null;
        } else {
            // make the argument list

            argList = new Class[args.length];

            for (int i = 0; i < args.length; i++) {
                argList[i] = args[i].getClass();

                // are we dealing with a primative (this is nasty)
                try {
                    /*
                     * You want to know how this works???
                     *
                     * A primative wrapper (eg java.lang.Integer) has a field called
                     * TYPE of class java.lang.Class. It is created by the JVM to
                     * represent a primative (eg int) in reflection.
                     * Success is things miraculously not going wrong, exceptions are
                     * thrown by the following (in order).
                     * No field called TYPE in argList[i]
                     * TYPE is not static (get(null) gets static fields)
                     * TYPE is not of java.lang.Class (initial cast)
                     */
                    argList[i] = (Class) (argList[i].getField("TYPE").get(null));
                } catch (Exception e) {
                    // an exception is thrown if the class is NOT a primitive so we
                    // don't make any action
                }
            }
        } // if args is null

        for (int i = 0; i < clnts.length; i++) {
            try {
                // fetch the method
                Method m = clnts[i].getClass().getMethod(name, argList);

                // invoke the method
                clnts[i] = m.invoke(clnts[i], args);
            } catch (Exception e) {
                // can't abort (action must be called on as many things as possible)
                e.printStackTrace();
            }
        }

        return clnts;
    }

} // AbstractMulticaster.class
