// RiskNetworkDialog.java

package uk.org.redfelineninja.risk.client;

import uk.org.redfelineninja.awt.*;

import java.awt.*;
import java.awt.event.*;

import javax.swing.*;
import javax.swing.event.*;
import javax.swing.border.*;

/**
 * This class implements the network dialog used to connect to Risk Arena
 * servers. It is a free-standing modeless dialog box. It should have a
 * RiskNetworkDialogListener added to it and then be set visible. For an
 * example in use refere to RiskClientFrame.java.
 */
public class RiskNetworkDialog extends JDialog {

    boolean newGame;

    public RiskNetworkDialog(JFrame parent) {
        this(parent, true);
    }

    public RiskNetworkDialog(JFrame parent, boolean n) {
        // we HAVE to use the ternary operator here, there is no choice
        super(parent, (n ? Words.New_Game : Words.Join_Game), true);

        newGame = n;

        JPanel container = new JPanel();
        container.setLayout(new ColumnLayout());

        JComponent network = buildNetworkPanel();
        network.setBorder(new TitledBorder(Words.Network_Options));
        container.add(network);

        JComponent player = buildPlayerPanel();
        player.setBorder(new TitledBorder(Words.Player_Options));
        container.add(player);

        // if we are starting a new game we must behave differently
        if (newGame) {
            JComponent gameOptions = buildGameOptionsPanel();
            gameOptions.setBorder(new TitledBorder(Words.Game_Options));
            container.add(gameOptions);
        } // if( newGame )

        JPanel buttons = new JPanel();
        buttons.setLayout(new FlowLayout(FlowLayout.CENTER));

        JButton cancel = new JButton(Words.Cancel);
        JButton ok = new JButton(Words.OK);

        cancel.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                doCancelPressed();
            }
        });

        ok.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                doOkPressed(e);
            }
        });

        buttons.add(ok);
        buttons.add(cancel);
        getRootPane().setDefaultButton(ok);

        container.add(buttons);

        getContentPane().add(container);
        pack();
    }

    // ACCESSORS (hooks with which to take data from the Dialog)

    public boolean getNewGameFlag() {
        return newGame;
    }
    public String getHostname() {
        return hostname.getText();
    }
    public int getPort() {
        return numericPort;
    }
    public String getGameName() {
        return gamename.getText();
    }

    public String getPlayerName() {
        return name.getText();
    }
    public Color getPlayerColor() {
        return colorButton.getBackground();
    }

    public boolean getPrivacyFlag() {
        return privacy.isSelected();
    }
    public boolean getRandomizeFlag() {
        return randomize.isSelected();
    }

    // ACTIONS

    private void doCancelPressed() {
        this.setVisible(false);
    }

    /**
     * Validate the numeric fields and if valid close the dialog.
     */
    private void doOkPressed(ActionEvent ae) {
        try {
            numericPort = Integer.parseInt(port.getText());

            // is port number valid
            if ((numericPort > 0) && (numericPort < 10000)) {
                this.setVisible(false);
                ActionEvent action = new ActionEvent(this, ActionEvent.ACTION_PERFORMED, "OkPressed", ae.getModifiers());
                fireActionPerformed(action);
            } else {
                // complain about the port number, continue
                JOptionPane.showMessageDialog(this, Words.portInWrongRange, Words.portNotValid, JOptionPane.ERROR_MESSAGE);
            }
        } catch (NumberFormatException e) {
            JOptionPane.showMessageDialog(this, Words.portNotNumeric, Words.portNotValid, JOptionPane.ERROR_MESSAGE);
        }
    } // doOkPressed( ActionEvent )

    // INTERFACE CODE STARTS HERE

    /**
     * List of listeners registered to this class.
     */
    private EventListenerList listenerList = new EventListenerList();

    public void addActionListener(ActionListener l) {
        listenerList.add(ActionListener.class, l);
    }

    public void removeActionListener(ActionListener l) {
        listenerList.remove(ActionListener.class, l);
    }

    private void fireActionPerformed(ActionEvent e) {
        // this is not null
        Object[] listeners = listenerList.getListenerList();

        for (int i = listeners.length - 2; i >= 0; i -= 2) {
            if (listeners[i] == ActionListener.class) {
                ((ActionListener) listeners[i + 1]).actionPerformed(e);
            } // if
        } // for
    } // fireActionPerformed( ActionEvent )

    /*
     * NETWORK PROPERTIES, these are the values the Dialog is trying to read
     * (in component form)
     */

    JTextField hostname, port, gamename, name;
    JCheckBox privacy, randomize;

    JButton colorButton;

    // having parsed port for validation keep in numeric form
    int numericPort;

    private void doColorButton() {
        Color c = JColorChooser.showDialog(this, Words.Select_Player_Color, colorButton.getBackground());

        if (c != null) {
            // colorIcon.setColor( c );
            colorButton.setBackground(c);
            colorButton.repaint();
        }
    } // doColorButton()

    private JComponent buildNetworkPanel() {
        JPanel network = new JPanel();

        network.setLayout(new LabeledPairLayout());

        network.add(new JLabel(Words.Hostname), "label");
        network.add(new JLabel(Words.Port), "label");
        network.add(new JLabel(Words.Game_Name), "label");

        // the port is just a port that sprang to mind for no reason
        hostname = new JTextField(Words.localhost);
        port = new JTextField("1644");
        gamename = new JTextField(Words.Risk);

        network.add(hostname, "field");
        network.add(port, "field");
        network.add(gamename, "field");

        return network;
    } // buildNetworkPanel()

    private JComponent buildPlayerPanel() {
        JPanel player = new JPanel();

        player.setLayout(new LabeledPairLayout());

        player.add(new JLabel(Words.Name), "label");
        player.add(new JLabel(Words.Color), "label");

        name = new JTextField(Words.Ysabelle);

        colorButton = new JButton(" ");

        // pick a random colour for the player to be
        // colorIcon = new ColorBlockIcon( 
        colorButton.setBackground(new Color((int) (Math.random() * 0x01000000) | 0xff000000));

        // colorButton = new JButton( colorIcon );

        // wire up actions here
        // best to interegate for state afterwards
        colorButton.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                doColorButton();
            }
        });

        player.add(name, "field");
        player.add(colorButton, "field");

        return player;
    }

    private JComponent buildGameOptionsPanel() {
        JPanel gameOptions = new JPanel();

        gameOptions.setLayout(new ColumnLayout());

        privacy = new JCheckBox(Words.Allow_Private_Talking);
        randomize = new JCheckBox(Words.Randomized_Start);

        randomize.setSelected(true);

        gameOptions.add(privacy);
        gameOptions.add(randomize);

        return gameOptions;
    } // buildGameOptionsPanel()
} // RiskNeworkingDialog.class
