// MapDefinition.java

package uk.org.redfelineninja.risk.io;

import uk.org.redfelineninja.util.*;
import uk.org.redfelineninja.risk.data.*;
import java.awt.Point;
import java.io.*;
import java.util.*;

/**
 * This class reads and parses Risk Map definition files.
 *
 * @version 0.9
 * @author Daniel Thompson, Feb. 1999
 */
public class MapDefinition {
    private Adjacency adj;
    private Hashtable seeds;
    private Hashtable continents;
    private String mapPath;
    private File imgFile = null;

    // this is a list of countries declared implicitly. It must be scanned when
    // the whole file has been read to ensure all implicit countries are made
    // explicit at some point
    private StringVector implicitCountries = new StringVector();

    private ContinentVector contIndex = new ContinentVector();

    // why have magic numbers unessessarily
    private final char quoteChar = '"';
    private final char[] quoteChars = { quoteChar };

    /**
     * 
     */
    public MapDefinition(File def) throws FileNotFoundException {
        // initialization
        adj = new Adjacency();
        seeds = new Hashtable();
        continents = new Hashtable();

        // set the map path
        mapPath = def.getParent();

        // open the file into a tokenizer
        StreamTokenizer tok = new StreamTokenizer(new BufferedReader(new FileReader(def)));

        // set up the tokenizer for our purposes (cf def. file format)
        tok.resetSyntax();
        tok.commentChar('#');
        tok.quoteChar(quoteChar);
        tok.wordChars('a', 'z');
        tok.wordChars('A', 'Z');
        tok.parseNumbers();
        tok.whitespaceChars(' ', ' ');
        tok.whitespaceChars('\t', '\t');
        tok.whitespaceChars('\n', '\n');
        tok.whitespaceChars('\r', '\r');
        tok.eolIsSignificant(true);

        // parse the file (will modify state variables)
        try {
            parseFile(tok);
            postCheckData();
        } catch (IOException e) {
            // an IO exception is such a crippling error that it kills the parse

            System.err.println("Serious internal error whilst parsing the map definition");
            e.printStackTrace();
        }
    }

    /**
     * Use this class to load the supplied preferences file. When called as an
     * application this method can be used to verify that the prefs file is
     * correct.
     */
    public static void main(String args[]) {
        if (args.length != 1) {
            System.err.println("usage: uk.ac.warwick.risk.io.MapDefinition <filename>");
            System.exit(3);
        }

        MapDefinition md = null;

        try {
            File f = new File(args[0]);
            md = new MapDefinition(f);
        } catch (FileNotFoundException e) {
            System.err.println(args[0] + ": file not found");
            System.exit(2);
        }

        System.out.println(md.getAdjacency().toString());
        System.out.println("\n\n");
        System.out.println(md.getSeeds().toString());
        System.out.println("\n\n");
        System.out.println(md.getContinents().toString());

        // uses enough AWT to start non daemon threads
        System.exit(0);
    } // main( String [] )

    /**
     * Parse the supplied Stream as a Risk definition file.
     *
     * @return exit code indicating success or failure
     */
    private int parseFile(StreamTokenizer tok) throws IOException {

        // go through the file line by line
        do {
            // get the token at start of line
            String cmd = Token.getText(tok, quoteChars);

            // determine of the token is a valid start of line
            if (cmd != null) {
                // we have a valid start of line check the type

                if (cmd.equals("image")) {
                    if (imgFile == null) {
                        // get the image (no sideeffects)
                        imgFile = getImageFile(tok);
                    } else if (!imgFile.exists()) {
                        // get the image (no sideeffects)
                        imgFile = getImageFile(tok);
                    } else {
                        // two images makes no sense
                        parseError(tok, "Map image set more than once");
                    }
                } else if (cmd.equals("continent")) {
                    // get the continent (has sideeffects)
                    getContinent(tok);
                } else {
                    // we have a country definition (has sideeffects)
                    getCountry(tok);
                }

            } else {
                if (!Token.getEOL(tok)) {
                    // we will ignore this line and report an error
                    parseError(tok, "Invalid first token");
                }
            }
        } while (tok.ttype != tok.TT_EOF);

        return 0;
    } // parseFile( StreamTokenizer )

    private File getImageFile(StreamTokenizer tok) throws IOException {
        // get and check the filename
        String filename = Token.getString(tok, quoteChars);
        if (filename == null) {
            parseError(tok, "Map image filename must be quoted string");
            return null;
        }

        // open the image
        File f;
        if ((mapPath == null) || (filename.indexOf(File.separator) == -1)) {
            // file name is the map directory
            f = new File(mapPath, filename);
        } else {
            // filename is absolute
            f = new File(filename);
        }

        if (!f.exists()) {
            parseError(tok, "Image " + filename + " does not exist");
        }

        // wind to eol
        tok.nextToken();

        if ((tok.ttype != tok.TT_EOL) && (tok.ttype != tok.TT_EOF)) {
            // too many arguments (parse error will wind to EOL)
            parseError(tok, "Spurious tokens after setting image");
        }

        return f;
    } // getImage( StreamTokenizer )

    private void getContinent(StreamTokenizer tok) throws IOException {
        // get and check continent
        String name = Token.getText(tok, quoteChars);
        if (name == null) {
            parseError(tok, "Name of continent must be text");
            return;
        }

        // get and check value
        Double value = Token.getNumber(tok);
        if (value == null) {
            parseError(tok, "Value of continent must be numeric");
            return;
        }

        // get and check tag
        Double tag = Token.getNumber(tok);
        if (tag == null) {
            parseError(tok, "Value of continent tag must be numeric");
            return;
        }

        Continent con = new Continent(name, value.intValue());
        if (continents.put(name, con) != null) {
            // we have a non fatal error (continents of same name declared)
            // we must continue after reporting the error because we have put a
            // new value into the hashtable. The vector must match up.
            parseError(tok, "Continent of that name already exists");
        }

        int t = tag.intValue();

        // grow contIndex if needed
        if (contIndex.size() <= t) {
            contIndex.setSize(t + 1);
        }

        contIndex.insertElementAt(con, t);
    } // getContinent( StreamTokenizer )

    private void getCountry(StreamTokenizer tok) throws IOException {
        // our name is the current token
        String country = tok.sval;

        // get and check the continent tag
        Double tag = Token.getNumber(tok);
        if (tag == null) {
            parseError(tok, "Value of continent tag must be numeric");
            return;
        }

        // add the country to the required data types
        if (!adj.contains(country)) {
            // must add country to matrix
            adj.addElement(country);
        } else {
            // must remove country from list of implicit countries
            implicitCountries.removeElement(country);
        }

        // add this country to its continent
        contIndex.continentAt(tag.intValue()).addCountry(country);

        while (!Token.getEOL(tok)) {
            // get and check next token
            String adjCountry = Token.getText(tok, quoteChars);
            if (adjCountry != null) {
                // we have a new adjacent country

                // if adj does not contain this country then introduce it implicitly
                if (!adj.contains(adjCountry)) {
                    adj.addElement(adjCountry);
                    implicitCountries.addElement(adjCountry);
                }

                // specify the adjacency
                adj.doubleLink(country, adjCountry);

                // break out to next token
                continue;
            }

            // we are expecting a pair of coordinates
            Character leftb = Token.getCharacter(tok, quoteChars);

            if (leftb == null) {
                // have failed to recognize token
                parseError(tok, "Expected a country name or a coord pair");

                // we have failed
                return;
            }

            // have we got a \ (continue on next line)
            if (leftb.charValue() == '\\') {
                if (Token.getEOL(tok)) {
                    // continue on the next line
                    continue;
                } else {
                    // \ must be at the end of a line
                    parseError(tok, "\\ must be the last character on a line");
                    return;
                }
            }

            // get the rest of the coord pair
            Double x = Token.getNumber(tok);
            Character comma = Token.getCharacter(tok, quoteChars);
            Double y = Token.getNumber(tok);
            Character rightb = Token.getCharacter(tok, quoteChars);

            if ((rightb == null) || (comma == null) || (x == null) || (y == null)) {
                // have failed to recognize token
                parseError(tok, "Expected a country name or a coord pair");

                // we have failed
                return;
            }

            // we are relaxed about what the characters actually are (people like
            // different brackets, who am i to argue)

            if (seeds.contains(country)) {
                // currently all but the first point is ignored. This is not the
                // first point so we break to the next token
                continue;
            }

            Point p = new Point(x.intValue(), y.intValue());
            seeds.put(country, p);
        }

        // ensure each country is linked to itself
        adj.link(country, country);
    } // getCountry( StreamTokenizer )

    private void postCheckData() throws IOException {
        if (implicitCountries.size() != 0) {
            StringBuffer sb = new StringBuffer();

            for (int i = 0; i < implicitCountries.size(); i++) {
                sb.append(implicitCountries.stringAt(i));
                sb.append(", ");
            }

            parseError(null, sb + " were declared implicitly");
        }
    }

    private void parseError(StreamTokenizer tok, String e) throws IOException {
        if (tok == null) {
            System.err.println("Error in map definition file");
            System.err.println(e);
        } else {
            System.err.println("Error in map definition file at line " + tok.lineno());
            System.err.println(e);

            // wind onto the next line
            while (!Token.getEOL(tok)) {
                tok.nextToken();
            }
        }
    } // parseError( String )

    public Adjacency getAdjacency() {
        return adj;
    }

    public Hashtable getSeeds() {
        return seeds;
    }

    public Hashtable getContinents() {
        return continents;
    }

    public File getImageFile() {
        return imgFile;
    }
}
