// RiskServerInterface.java

package uk.org.redfelineninja.risk.net;

import uk.org.redfelineninja.risk.data.*;

import java.io.*;

/**
 * All actions that can be performed on the server.
 */
public interface RiskServerInterface {

    /**
     * Open a connection to the server.
     */
    public void openConnection(String hostname, int port, Player player) throws IOException;

    /**
     * Close the current connection. This will cause the player to leave the
     * game and this is impolite. The server however is not in a position to
     * enforce good manners.
     */
    public void closeConnection() throws IOException;

    /**
     * Test whether or not we are connected to a server.
     */
    public boolean isConnected();

    /**
     * Start a new game of risk with the supplied name
     *
     * @param player details of the player starting the game
     * @param name the name that the game will be identified with
     * @param random start with randomly owned countries
     * @param privacy permit private talking on the message system
     */
    public void newGame(String gamename, boolean random, boolean privacy);

    /**
     * Join an existing game of Risk
     *
     * @param player details of the player joining the game
     */
    public void joinGame(String gamename);

    /**
     * Start the game of risk. After as many players a desired have joined the
     * game then call this method.
     */
    public void startGame();

    /**
     * Stop the current game of risk. This should be considered impolite but at
     * times may be unavaoidable. Only the person who started the game is
     * permitted to stop it.
     *
     * @param reason the reason for the game being halted
     */
    public void stopGame(String reason);

    /**
     * Send a message to all players.
     *
     * @param message the message
     */
    public void message(String message);

    /**
     * Send a message to the named player. This may be forbidden by the server
     * in some games.
     *
     * @param message the message
     * @param player the player to send the message to
     */
    public void message(String message, Player player);

    /**
     * Inform the server that a country has been clicked on. Since text is used
     * to identify countries, it is helpful if the server and client use
     * the same set up files.
     *
     * @param country string giving the country that was clicked on.
     */
    public void countryClicked(String country);

    /**
     * Exchange the supplied cards for armies. Subject to checking by the server
     * obviously.
     *
     * @param cards the card to be exchanged
     */
    public void cards(Card[] cards);

    /**
     * Change into fortify mode.
     */
    public void fortify();

    /**
     * Complete the current turn.
     */
    public void finishTurn();
} // RiskServerInterface.interface
