// Robot.java

package uk.org.redfelineninja.risk.robot;

import uk.org.redfelineninja.risk.client.RiskServerHandler;
import uk.org.redfelineninja.risk.data.Player;
import uk.org.redfelineninja.risk.net.RiskServerInterface;
import uk.org.redfelineninja.risk.net.Protocol;

import java.io.*;
import java.util.Random;

/**
 * This is the superclass from which all robot players should be derived.
 *
 * The intructions for using this class can be found in the user manual.
 */
public abstract class Robot {
    protected GameState game;
    protected RiskServerInterface server;

    protected static Random random = new Random();

    protected Player me;

    protected String defaultName = "Robot";

    public Robot() {} // constuctor()

    public void setPlayer(Player p) {
        me = p;
    } // setPlayer( Player )

    public void setServer(RiskServerInterface s) {
        server = s;
    } // setServer( ... )

    public void setGame(GameState g) {
        game = g;
    } // setGame( GameState )

    public Player getIdentity() {
        return me;
    } // getIdentity()

    public abstract Move bestMove(int mode);

    /**
     * Connect the robot to the server.
     *
     * This method should be called from the robots <code>main</code> method
     * after any relavent argument processing.
     */
    public static void main(Robot r, String args[]) {
        RiskServerInterface server;

        // command line details
        String name = r.defaultName;
        String hostname = "localhost";
        int port = 1644;
        boolean block = true;

        // get the command line arguments
        try {
            for (int i = 0; i < args.length; i++) {
                String arg = args[i].substring(2);

                if (arg.equals("name")) {
                    name = args[++i];
                } else if (arg.equals("hostname")) {
                    hostname = args[++i];
                } else if (arg.equals("port")) {
                    port = Integer.parseInt(args[++i]);
                } else if (arg.equals("noblock")) {
                    block = false;
                } else {
                    System.err.println(
                        "Bad command line arguments\nUsage: Robot [--name <player_name>] [--hostname <hostname>] [--port <port>]");
                    System.exit(1);
                }
            } // for( every argument )
        } catch (Exception e) {
            System.err.println(
                "Bad command line arguments\nUsage: Robot [--name <player_name>] [--hostname <hostname>] [--port <port>]");
            System.exit(1);
        }

        // set up the robot name
        r.setPlayer(new Player(name, random.nextInt(), true));

        // attempt to start a game
        try {
            File f = new File("maps" + File.separator + "risk.map");
            GameState g = new GameState(f, r);

            server = new RiskServerHandler(g);
            g.setServer(server);
            r.setServer(server);
            r.setGame(g);

            server.openConnection(hostname, port, r.getIdentity());
            server.message("/" + Protocol.sRobot + " " + Protocol.sOn);
            server.joinGame("risk.map");

        } catch (Exception e) {
            e.printStackTrace();
            return;
        }

        BufferedReader in = new BufferedReader(new InputStreamReader(System.in));

        // Do not exit (Server handler is a daemon thread
        while (block) {
            try {
                Thread.sleep(10000);
            } catch (InterruptedException e) {}

        } // for ever

    } // main( Player, String, int )

} // Robot.java
