// RiskClientWriter.java

package uk.org.redfelineninja.risk.server;

import uk.org.redfelineninja.risk.data.*;
import uk.org.redfelineninja.risk.net.Protocol;
import uk.org.redfelineninja.risk.net.RiskClientInterface;

import java.io.*;
import java.util.Hashtable;

/**
 * This class provides the network output hooks required by the server. 
 * The RiskServerInterface proved to be insufficient so we have extended from
 * writer as well.
 */
public class RiskClientWriter extends PrintWriter implements RiskClientInterface {
    // used to ensure privacy for setCards
    private Player activePlayer;

    // used to look up players private output streams
    private Hashtable playerRegistry; // (Player) -> (PrintWriter)

    private static final String prompt = Protocol.sPrompt + String.valueOf(Protocol.command);

    public RiskClientWriter(Writer network, Hashtable pR) {
        super(network);
        playerRegistry = pR;
    } // this( Writer )

    // SERVER UTILS (used only by the RiskEngine -> package access)

    /**
     * Report an error only to the active player.
     */
    public void error(String err) {
        ((PrintWriter) playerRegistry.get(activePlayer)).println(Protocol.sPrompt + String.valueOf(Protocol.warning) + err);
    }

    // INTERFACE IMPLEMENTATION

    /**
     * Called when a new player has joined the game the client is waiting to play
     */
    public void newPlayer(Player player) {
        println(prompt + Protocol.cNewPlayer + " \"" + player.name + "\" " + player.color.getRGB());
    }

    /**
     * Called whenever a new spectator has joined the game.
     *
     * This can be ignored unless the client has a talker that supports private
     * talking. Robots could communicate their internal state like this if
     * desired.
     */
    public void newSpectator(Player player) {
        println(prompt + Protocol.cNewSpectator + " \"" + player.name + "\"");
    }

    /**
     * Called when a player leaves the game (or if his connection is timed out).
     *
     * Leaving mid-game is anti-social but it will happen.
     */
    public void lostPlayer(Player player) {
        println(prompt + Protocol.cLostPlayer + " \"" + player.name + "\"");
    }

    // GAME CONTROL

    /**
     * Called when the games owner decides to stop waiting for new players and
     * start playing.
     */
    public void newGame(Player[] players) {
        StringBuffer command = new StringBuffer();

        command.append(prompt).append(Protocol.cNewGame);

        for (int i = 0; i < players.length; i++) {
            command.append(" \"").append(players[i].name).append("\"");
        }

        println(command.toString());
    }

    /**
     * Called when the game is won.
     */
    public void gameOver(Player victor) {
        println(prompt + Protocol.cGameOver + " \"" + victor.name + "\"");
    }

    /**
     * Called when the game is stopped by the server.
     */
    public void gameOver(Player player, String reason) {
        println(prompt + Protocol.cGameOver + " \"" + player.name + "\" \"" + reason + "\"");
    }

    // MESSAGING

    /**
     * Message sent over the chat system to all clients.
     */
    public void message(String message) {
        println(message);
    }

    // TURN CONTROL

    /**
     * It is player's turn
     */
    public void setTurn(Player player) {
        // store this for use in setCards
        activePlayer = player;

        println(prompt + Protocol.cSetTurn + " \"" + player.name + "\"");
    }

    // DISPLAY CONTROL

    /**
     * Display the supplied string in the clients status line.
     */
    public void showStatus(String status) {
        println(prompt + Protocol.cShowStatus + " \"" + status + "\"");
    }

    /**
     * Set the number of armies in a country.
     */
    public void setArmies(String country, int a) {
        println(prompt + Protocol.cSetArmies + " \"" + country + "\" " + a);

    }

    /**
     * Set the owner of the country to the player.
     */
    public void setOwner(String country, Player player) {
        println(prompt + Protocol.cSetOwner + " \"" + country + "\" \"" + player.name + "\"");
    }

    /**
     * Set the selection to the country.
     */
    public void select(String country) {
        if (country == null) {
            println(prompt + Protocol.cSelect);
        } else {
            println(prompt + Protocol.cSelect + " \"" + country + "\"");
        }
    } // select( Country )

    /**
     * Supply the current state of the dice to those it interests.
     */
    public void setDice(int[] die) {
        StringBuffer command = new StringBuffer();

        command.append(prompt).append(Protocol.cSetDice);

        for (int i = 0; i < die.length; i++) {
            command.append(" ").append(die[i]);
        }

        println(command);
    }

    /**
     * Tell the client the cards that the player has.
     *
     * Clients can connect to the server with two connections if there are two
     * players at that client.
     */
    public void setCards(Card[] cards) {
        StringBuffer command = new StringBuffer();

        command.append(prompt).append(Protocol.cNewGame);

        for (int i = 0; i < cards.length; i++) {
            command.append(" \"").append(cards[i].country).append("\"");
        }

        ((PrintWriter) playerRegistry.get(activePlayer)).println(command.toString());
    }

} // RiskClientWriter.class
